package org.ibex.arenaj;
import soot.*;
import soot.jimple.*;
import soot.util.*;
import java.io.*;
import java.util.*;

public class Main extends SceneTransformer {    

    public static final int initialSize = 1000;

    private static Main instance = new Main();
    private Main() { }
    public static Main v() { return instance; }
   
    public static void main(String[] args) throws Exception {
        if(args.length == 0) {
            System.out.println("Syntax: java " + v().getClass().getName() + " [soot options]");
            System.exit(0);
        }
        PackManager.v().getPack("wjtp").add(new Transform("wjtp.tx", Main.v()));
        String[] args2 = new String[args.length + 14];
        System.arraycopy(args, 0, args2, 0, args.length-1);
        String sootcp =
            System.getProperty("java.class.path") + 
            File.pathSeparator +
            System.getProperty("sun.boot.class.path");
        String extDirs = System.getProperty("java.ext.dirs");
        if (extDirs != null) {
            StringTokenizer st = new StringTokenizer(extDirs, File.pathSeparatorChar+"");
            while(st.hasMoreTokens()) {
                String goo = st.nextToken();
                System.out.println("goo " + goo);
                File f = new File(goo);
                if (!f.exists() || !f.isDirectory()) continue;
                String[] jars = f.list();
                for(int i=0; i<jars.length; i++)
                    sootcp += File.pathSeparator + f.getPath() + File.separatorChar + jars[i];
            }
        }
        args2[args.length - 1] = "-w";
        args2[args.length + 0] = "-p";
        args2[args.length + 1] = "cg";
        args2[args.length + 2] = "enabled:false";
        args2[args.length + 3] = "-p";
        args2[args.length + 4] = "wjap";
        args2[args.length + 5] = "enabled:false";
        args2[args.length + 6] = "-p";
        args2[args.length + 7] = "jtp";
        args2[args.length + 8] = "enabled:false";
        //args2[args.length + 9] = "-keep-line-number";
        args2[args.length + 9] = "-allow-phantom-refs";
        args2[args.length + 10] = "-allow-phantom-refs";
        args2[args.length + 11] = "-cp";
        args2[args.length + 12] = sootcp;
        System.out.println("sootcp => " + sootcp);
        args2[args.length + 13] = args[args.length-1];
        for(int i=0; i<args2.length; i++) System.out.println("args2["+i+"] = " + args2[i]);
        soot.Main.main(args2);
    }

    // The main loop //////////////////////////////////////////////////////////////////////////////

    public void internalTransform(String phaseName, Map options) {
        System.out.println("begun");
        init();
        List ac = new LinkedList();
        for(Iterator it = Scene.v().getClasses().iterator(); it.hasNext();) {
            SootClass sc = (SootClass)it.next();
            sc.setLibraryClass();
        }
        for(Iterator it = Scene.v().getClasses().iterator(); it.hasNext();) {
            SootClass sc = (SootClass)it.next();
            if (implementsGladiator(sc)) {
                System.out.println("  application class: " + sc.getName());
                System.out.println("  application class: " + getArenaForGladiator(sc).getName());
                sc.setApplicationClass();
                getArenaForGladiator(sc).setApplicationClass();
                ac.add(sc);
                ac.add(getArenaForGladiator(sc));
                /*
                for(Iterator i2 = sc.getMethods().iterator(); i2.hasNext();) {
                    SootMethod m = (SootMethod)i2.next();
                    if (m.isConcrete()) m.retrieveActiveBody();
                }
                for(Iterator i2 = getArenaForGladiator(sc).getMethods().iterator(); i2.hasNext();) {
                    SootMethod m = (SootMethod)i2.next();
                    if (m.isConcrete()) m.retrieveActiveBody();
                }
                */
            }
        }
        for(Iterator it = ac.iterator(); it.hasNext();) {
            SootClass sc = (SootClass)it.next();
            if (sc.isInterface() || !implementsGladiator(sc)) continue;
            processGladiatorClass(sc);
        }
        for(Iterator it = ac.iterator(); it.hasNext();) {
            SootClass sc = (SootClass)it.next();
            if (sc.isInterface()) continue;
            processClass(sc);
        }
    }

    // Initializers //////////////////////////////////////////////////////////////////////////////

    public Type objectType;
    public List arrayCopyTypes = new LinkedList();
    public SootMethodRef arrayCopy;
    public void init() {
        objectType = Scene.v().getSootClass("java.lang.Object").getType();
        arrayCopyTypes.add(objectType);
        arrayCopyTypes.add(IntType.v());
        arrayCopyTypes.add(objectType);
        arrayCopyTypes.add(IntType.v());
        arrayCopyTypes.add(IntType.v());
        arrayCopy = 
            Scene.v().makeMethodRef(Scene.v().getSootClass("java.lang.System"),
                                    "arraycopy",
                                    arrayCopyTypes,
                                    VoidType.v(),
                                    true);
    }

    // Helpers //////////////////////////////////////////////////////////////////////////////

    private int tfr = 0;
    public Local newLocal(Body b, Type t) {
        Local l = Jimple.v().newLocal("tmpRef" + (tfr++), t);
        b.getLocals().add(l);
        return l;
    }

    public Local viaLocal(Value v, Body b, Unit u) {
        Local l = newLocal(b, v.getType());
        if (v instanceof IdentityRef) b.getUnits().insertBefore(Jimple.v().newIdentityStmt(l, v), u);
        else b.getUnits().insertBefore(Jimple.v().newAssignStmt(l, v), u);
        return l;
    }

    boolean          implementsGladiator(SootClass c)  { return c.implementsInterface("org.ibex.arenaj.Gladiator"); }
    boolean          implementsGladiator(Type t) {return (t instanceof RefType)&&implementsGladiator(((RefType)t).getSootClass());}
    String           getArenaName(SootClass sc) { return sc.getName().substring(0, sc.getName().lastIndexOf('$')); }
    String           getGladiatorName(SootClass sc) { return sc.getName().substring(sc.getName().lastIndexOf('$')+1); }
    SootClass        getArenaForGladiator(SootClass sc) { return Scene.v().getSootClass(getArenaName(sc)); }
    InstanceFieldRef newIFR(Body b, SootFieldRef fr) { return Jimple.v().newInstanceFieldRef(b.getThisLocal(), fr); }
    void             assign(Body b, Value left, Value right) { b.getUnits().add(Jimple.v().newAssignStmt(left, right)); }
    void             assign(Body b, Value l, Value r, Stmt w) { b.getUnits().insertBefore(Jimple.v().newAssignStmt(l, r), w); }
    Type             getSliceElementType(Type t) { return implementsGladiator(t) ? IntType.v() : t; }
    SootFieldRef     getSliceForField(SootField f)    { return getSliceForField(f.makeRef()); }
    SootFieldRef     getSliceForField(SootFieldRef f) {
        return Scene.v().makeFieldRef(getArenaForGladiator(f.declaringClass()),
                                      getGladiatorName(f.declaringClass())+"$$"+f.name(),
                                      getSliceElementType(f.type()).makeArrayType(),
                                      false);
    }


    // Operations performed on the Gladiator class //////////////////////////////////////////////////////////////////////////

    public void processGladiatorClass(SootClass sc) {
        SootClass mc = getArenaForGladiator(sc);
        String subname = sc.getShortName().substring(sc.getShortName().lastIndexOf('$')+1);
        String incFuncName = subname + "$$inc";
        SootMethod method = new SootMethod(incFuncName, new LinkedList(), IntType.v(), Modifier.PRIVATE, new LinkedList());

        mc.setApplicationClass();
        SootMethod initMethod = mc.getMethod("<init>", new LinkedList());
        if (!initMethod.hasActiveBody()) initMethod.setActiveBody((JimpleBody)Jimple.v().newBody(initMethod));
        Body mcInitBody = initMethod.getActiveBody();

        mc.addMethod(method);
        Body body = Jimple.v().newBody(method);
        method.setActiveBody(body);
        ((JimpleBody)body).insertIdentityStmts();
        Local l  =  newLocal(body, IntType.v());
        Local l2 =  newLocal(body, IntType.v());
        Local l3 =  newLocal(body, IntType.v());
        Local l4 =  newLocal(body, IntType.v());

        Stmt where = ((JimpleBody)mcInitBody).getFirstNonIdentityStmt();
        SootField maxField = new SootField(subname + "$$max", IntType.v());
        mc.addField(maxField);
        assign(mcInitBody, newIFR(mcInitBody, maxField.makeRef()), IntConstant.v(initialSize), where);

        SootField sfr = new SootField(subname + "$$size", IntType.v());
        mc.addField(sfr);
        assign(mcInitBody, newIFR(mcInitBody, sfr.makeRef()), IntConstant.v(0), where);
        
        assign(body, l,                                 newIFR(body, sfr.makeRef()));
        assign(body, l2,                                Jimple.v().newAddExpr(l, IntConstant.v(1)));
        assign(body, newIFR(body, sfr.makeRef()),       l2);
        assign(body, l3,                                newIFR(body, maxField.makeRef()));

        Stmt returnStmt = Jimple.v().newReturnStmt(l2);
        body.getUnits().add(Jimple.v().newIfStmt(Jimple.v().newLtExpr(l2, l3), returnStmt));

        assign(body,  l4,                               Jimple.v().newShlExpr(l3, IntConstant.v(1)));
        assign(body,  newIFR(body, maxField.makeRef()), l4);

        for(Iterator it = sc.getFields().iterator(); it.hasNext();) {
            SootField f = (SootField)it.next();
            Type t      = getSliceElementType(f.getType());
            mc.addField(f = new SootField(getGladiatorName(sc) + "$$" + f.getName(), t.makeArrayType(), 0));

            Expr newArr = Jimple.v().newNewArrayExpr(t, IntConstant.v(initialSize));
            Local newArrLocal = newLocal(mcInitBody, f.getType());
            mcInitBody.getUnits().addFirst(Jimple.v().newAssignStmt(newIFR(mcInitBody, f.makeRef()), newArrLocal));
            mcInitBody.getUnits().addFirst(Jimple.v().newAssignStmt(newArrLocal, newArr));

            Local ll0 = newLocal(body, f.getType());
            Local ll = newLocal(body, f.getType());
            assign(body, ll0, newIFR(body,  f.makeRef()));
            assign(body, ll,  Jimple.v().newNewArrayExpr(t, l4));

            List args = new LinkedList();
            args.add(ll0);
            args.add(IntConstant.v(0));
            args.add(ll);
            args.add(IntConstant.v(0));
            args.add(l3);
            body.getUnits().add(Jimple.v().newInvokeStmt(Jimple.v().newStaticInvokeExpr(arrayCopy, args)));
            assign(body, newIFR(body,  f.makeRef()), ll);
        }

        for(Iterator it = sc.getMethods().iterator(); it.hasNext();) {
            SootMethod m = (SootMethod)it.next();
            if (!m.isConcrete()) continue;
            boolean doremove = true;
            Body mbody = m.getActiveBody();
            if (implementsGladiator(m.getDeclaringClass()) && m.getName().equals("<init>")) {
                doremove = false;
                SootClass c = m.getDeclaringClass();
                String name = /*c.getShortName().substring(c.getShortName().lastIndexOf('$')+1) + */"$init";
                List li = m.getParameterTypes();
                c.removeMethod(m);
                SootMethod nm = new SootMethod(name, li, implementsGladiator(m.getReturnType()) ? IntType.v() : m.getReturnType());
                JimpleBody bod = Jimple.v().newBody(nm);
                bod.importBodyContentsFrom(m.getActiveBody());
                nm.setActiveBody(bod);
                m = nm;
                mbody = bod;
                mbody.getUnits().remove(((JimpleBody)mbody).getFirstNonIdentityStmt());
                mbody.getUnits().remove(((JimpleBody)mbody).getFirstNonIdentityStmt());
            } else {
                System.out.println("examining " + sc.getName() + "." + m.getSignature());
                m.retrieveActiveBody();
            }
            if (m.isStatic()) continue;

            String name = sc.getShortName().substring(sc.getShortName().lastIndexOf('$')+1) + "$$" + m.getName();
            List list = new LinkedList();
            list.addAll(m.getParameterTypes());
            list.add(IntType.v());
            for(Iterator i = list.iterator(); i.hasNext();) System.out.println(i.next());
            SootMethod m2 = new SootMethod(name, list, m.getReturnType());
            getArenaForGladiator(sc).addMethod(m2);

            JimpleBody ab = (JimpleBody)Jimple.v().newBody(m2);
            ab.importBodyContentsFrom(mbody);
            m2.setActiveBody(ab);
            Local loc = Jimple.v().newLocal("tmpRef" + (tfr++), getArenaForGladiator(sc).getType());
            ab.getLocals().add(loc);
            // FIXME: insert assignment to this

            for(Iterator z = ab.getLocals().iterator(); z.hasNext();) {
                Local loc2 = (Local)z.next();
                if (implementsGladiator(loc2.getType())) {
                    loc2.setType(IntType.v());
                }
            }

            Chain units = ab.getUnits();
            boolean touched = false;
            Local loc0 = Jimple.v().newLocal("tmpRef" + (tfr++), getArenaForGladiator(sc).getType());
            ab.getLocals().add(loc0);
            for(Iterator stmtIt = units.snapshotIterator(); stmtIt.hasNext();) {
                Stmt s = (Stmt) stmtIt.next();
                if (s instanceof IdentityStmt) {
                    IdentityStmt is = (IdentityStmt)s;
                    Local left = (Local)is.getLeftOp();
                    if (is.getRightOp() instanceof ThisRef) {
                        left.setType(IntType.v());
                        is.getRightOpBox().setValue(Jimple.v().newParameterRef(IntType.v(), m.getParameterCount()));
                        if (!touched) {
                            units.addFirst(Jimple.v().newIdentityStmt(loc0, Jimple.v().newThisRef(getArenaForGladiator(sc).getType())));
                            touched = true;
                        }
                    }
                }

                for(Iterator i = s.getUseAndDefBoxes().iterator(); i.hasNext();) {
                    Object o = i.next();
                    if (o instanceof ValueBox) {
                        ValueBox vb = (ValueBox)o;
                        o = vb.getValue();
                        /*
                        if (o instanceof Local && implementsGladiator(((Local)o).getType())) {
                            System.out.println("thunking");
                            vb.setValue(loc0);
                        }
                        */
                        if (vb.getValue() instanceof ThisRef) {
                            System.out.println(s);
                            vb.setValue(loc);
                        }
                    }
                }

            }
            
            if (doremove) sc.removeMethod(m);

        }
        body.getUnits().add(returnStmt);
    }

    // Operations performed on all classes ////////////////////////////////////////////////////////////////////////////

    public void processClass(SootClass c) {
        for(Iterator it = c.getFields().iterator(); it.hasNext();) {
            SootField f = (SootField)it.next();
            Type t = f.getType();
            if (t instanceof RefType) {
                RefType rt = (RefType)t;
                SootClass sc = rt.getSootClass();
                if (implementsGladiator(sc)) f.setType(IntType.v());
            } else if (t instanceof ArrayType) {
                ArrayType at = (ArrayType)t;
                t = at.getElementType();
                if (!(t instanceof RefType)) continue;                
                RefType rt = (RefType)t;
                SootClass sc = rt.getSootClass();
                if (implementsGladiator(sc)) f.setType(IntType.v().makeArrayType());
            }
        }

        List list = new LinkedList(); list.addAll(c.getMethods());
        for(Iterator it = list.iterator(); it.hasNext();) {
            SootMethod m = (SootMethod)it.next();
            Body b = null;
            if (m.getName().endsWith("$$inc")) continue;
            if (m.hasActiveBody()) b = processBody(m.getActiveBody(), c, m);
            List l2 = new LinkedList();
            List l = m.getParameterTypes();
            for(Iterator it2 = l.iterator(); it2.hasNext();) {
                Type t = (Type)it2.next();
                l2.add(implementsGladiator(t) ? IntType.v() : t);
            }
            Type t = m.getReturnType();
            if (implementsGladiator(t)) {
                t = IntType.v();
                if (m.hasActiveBody()) {
                    Body bod = m.getActiveBody();
                    for(Iterator stmtIt = bod.getUnits().snapshotIterator(); stmtIt.hasNext();) {
                        Stmt s = (Stmt) stmtIt.next();
                        if (s instanceof ReturnStmt) {
                            if (((ReturnStmt)s).getOp().getType() instanceof NullType) {
                                ((ReturnStmt)s).getOpBox().setValue(IntConstant.v(-1));
                            }
                        }
                    }
                }
            }
            String name = m.getName();
            SootMethod meth = new SootMethod(name, l2, implementsGladiator(t) ? IntType.v() : t, m.getModifiers());
            if (b != null) {
                JimpleBody b2 = Jimple.v().newBody(meth);
                b2.importBodyContentsFrom(b);
                meth.setActiveBody(b2);
            }
            c.removeMethod(m);
            c.addMethod(meth);
        }

    }

    protected Body processBody(Body body, SootClass ownerClass, SootMethod smeth) {
        if (body instanceof JimpleBody) {
            JimpleBody b2 = Jimple.v().newBody(smeth);
            b2.importBodyContentsFrom(body);
            body = b2;
        }
        Chain units = body.getUnits();
        System.out.println("processBody("+body.getMethod()+")");
        for(Iterator it = body.getLocals().snapshotIterator(); it.hasNext();) {
            Local l = (Local)it.next();
            if (implementsGladiator(l.getType())) l.setType(IntType.v());
        }
        for(int qq=0; qq<2; qq++) for(Iterator stmtIt = units.snapshotIterator(); stmtIt.hasNext();) {
            Stmt s = (Stmt) stmtIt.next();
            if (s instanceof DefinitionStmt) {
                DefinitionStmt ds = (DefinitionStmt)s;
                if (ds.getLeftOp().getType() instanceof PrimType && ds.getRightOp().getType() instanceof NullType) {
                    ds.getRightOpBox().setValue(IntConstant.v(-1));
                }
            }
            if (implementsGladiator(smeth.getReturnType()) && s instanceof ReturnStmt)
                if (((ReturnStmt)s).getOp().getType() instanceof NullType)
                    ((ReturnStmt)s).getOpBox().setValue(IntConstant.v(-1));
            List l = s.getUseAndDefBoxes();
            for(Iterator it = l.iterator(); it.hasNext();) {
                Object o = it.next();
                if (o instanceof ValueBox) {
                    ValueBox vb = (ValueBox)o;
                    Value v = vb.getValue();
                    
                    if (v instanceof BinopExpr) {
                        BinopExpr boe = (BinopExpr)v;
                        Type t1 = boe.getOp1().getType();
                        Type t2 = boe.getOp2().getType();
                        if (t1 instanceof PrimType && t2 instanceof NullType) boe.setOp2(IntConstant.v(-1));
                        if (t2 instanceof PrimType && t1 instanceof NullType) boe.setOp1(IntConstant.v(-1));
                    }

                    if (v instanceof NewExpr) {
                        NewExpr ne = (NewExpr)v;
                        if (implementsGladiator(ne.getBaseType())) {
                            SootClass sc = ((RefType)ne.getBaseType()).getSootClass();
                            SootClass mc = getArenaForGladiator(sc);
                            String incFuncName = sc.getShortName().substring(sc.getShortName().lastIndexOf('$')+1) + "$$inc";
                            SootMethodRef smr = Scene.v().makeMethodRef(mc, incFuncName, new LinkedList(), IntType.v(), false);
                            Expr invokeExpr = Jimple.v().newSpecialInvokeExpr(body.getThisLocal(), smr);
                            Local ll = viaLocal(invokeExpr, body, s);
                            vb.setValue(ll);
                            v = ll;
                            continue;
                        } 
                    } else

                    if (v instanceof InvokeExpr) {
                        InvokeExpr ie = (InvokeExpr)v;
                        SootMethodRef mr = ie.getMethodRef();
                        String name = mr.name();
                        if (v instanceof InstanceInvokeExpr && implementsGladiator(mr.declaringClass())) {
                            InstanceInvokeExpr iie = (InstanceInvokeExpr)v;
                            List li = new LinkedList();
                            li.addAll(iie.getArgs());
                            LinkedList pl = new LinkedList();
                            for(Iterator it2 = mr.parameterTypes().iterator(); it2.hasNext();) {
                                Type t = (Type)it2.next();
                                pl.add(implementsGladiator(t) ? IntType.v() : t);
                            }
                            if (mr.name().equals("<init>") && implementsGladiator(mr.declaringClass())) {
                                name = "$init";
                                //li.remove(0);
                                //pl.remove(0);
                            }
                            pl.add(IntType.v());
                            li.add(viaLocal(iie.getBase(),body,s));
                            SootClass sc = mr.declaringClass();
                            name = sc.getShortName().substring(sc.getShortName().lastIndexOf('$')+1) + "$$" + name;
                            mr = Scene.v().makeMethodRef(getArenaForGladiator(sc),
                                                         name,
                                                         pl,
                                                         implementsGladiator(mr.returnType()) ? IntType.v() : mr.returnType(),
                                                         false);
                            ie = Jimple.v().newVirtualInvokeExpr(body.getThisLocal(), mr, li);
                            vb.setValue(v = ie);

                        } else if (!(v instanceof StaticInvokeExpr)) {
                            List l0 = mr.parameterTypes();
                            List l2 = new LinkedList();
                            for(Iterator it2 = l0.iterator(); it2.hasNext();) {
                                Type t = (Type)it2.next();
                                l2.add(implementsGladiator(t) ? IntType.v() : t);
                            }
                            mr = Scene.v().makeMethodRef(mr.declaringClass(),
                                                         mr.name(),
                                                         l2,
                                                         implementsGladiator(mr.returnType()) ? IntType.v() : mr.returnType(),
                                                         mr.isStatic());
                            ie.setMethodRef(mr);
                            vb.setValue(v = ie);                            
                        }

                        for(int i=0; i<ie.getArgCount(); i++) {
                            ValueBox b = ie.getArgBox(i);
                            Value val = b.getValue();
                            if (mr.parameterType(i) instanceof RefType && val.getType() instanceof PrimType) {
                                SootClass intClass = Scene.v().getSootClass("java.lang.Integer");
                                List typelist = new LinkedList();
                                typelist.add(IntType.v());
                                SootMethod intMethod = intClass.getMethod("<init>", typelist);
                                Local loc = viaLocal(Jimple.v().newNewExpr(RefType.v(intClass)), body, s);
                                List list = new LinkedList();
                                list.add(val);
                                units.insertBefore(Jimple.v().newInvokeStmt(Jimple.v().newSpecialInvokeExpr(loc,
                                                                                                            intMethod.makeRef(),
                                                                                                            list)),
                                                   s);
                                b.setValue(loc);
                            }
                            if (val != null && val.getType() instanceof NullType && mr.parameterType(i) instanceof IntType) {
                                b.setValue(IntConstant.v(-1));
                            }
                        }


                    } else if (v instanceof CastExpr) {
                        CastExpr ce = (CastExpr)v;
                        if (implementsGladiator(ce.getCastType())) {
                            SootClass mc = getArenaForGladiator(((RefType)ce.getCastType()).getSootClass());
                            SootClass ic = Scene.v().getSootClass("java.lang.Integer");
                            ce.setCastType(ic.getType());

                            Local l1 = Jimple.v().newLocal("tmpRef" + (tfr++), ic.getType()); body.getLocals().add(l1);
                            Local l2 = Jimple.v().newLocal("tmpRef" + (tfr++), IntType.v()); body.getLocals().add(l2);

                            Stmt s2 = Jimple.v().newAssignStmt(l1, Jimple.v().newCastExpr(ce.getOp(), ic.getType()));
                            body.getUnits().insertBefore(s2, s);

                            Stmt isNull = Jimple.v().newAssignStmt(l2, IntConstant.v(-1));
                            body.getUnits().insertAfter(isNull, s2);

                            Stmt ifStmt = Jimple.v().newIfStmt(Jimple.v().newEqExpr(l1, NullConstant.v()), s);
                            body.getUnits().insertAfter(ifStmt, isNull);

                            SootMethodRef mr = Scene.v().makeMethodRef(ic, "intValue", new LinkedList(), IntType.v(), false);
                            Stmt isNotNull =
                                Jimple.v().newAssignStmt(l2, Jimple.v().newVirtualInvokeExpr(l1, mr, new LinkedList()));
                            body.getUnits().insertAfter(isNotNull, ifStmt);

                            vb.setValue(l2);
                            qq = 0;  // ???
                            break;
                        }

                    } else if (v instanceof InstanceFieldRef) {
                        InstanceFieldRef ifr = (InstanceFieldRef)v;
                        SootFieldRef fr = ifr.getFieldRef();
                        Type t = fr.type();
                        if (implementsGladiator(fr.declaringClass()) && fr.name().equals("this$0")) {
                            System.out.println("******************");
                            vb.setValue(body.getThisLocal());
                            System.out.println("s is " + s);
                        } else if (implementsGladiator(fr.declaringClass())) {
                            SootClass mc = getArenaForGladiator(fr.declaringClass());
                            InstanceFieldRef sfr = newIFR(body,  getSliceForField(fr));
                            vb.setValue(Jimple.v().newArrayRef(viaLocal(sfr, body, s), ifr.getBase()));
                        }
                        if ((t instanceof RefType) && implementsGladiator(((RefType)t).getSootClass())) {
                            SootClass tc = ((RefType)t).getSootClass();
                            SootClass mc = getArenaForGladiator(tc);
                            ifr.setFieldRef(Scene.v().makeFieldRef(mc, fr.name(), IntType.v(), false));
                        }
                    }

                }
            }
        }
        body.validate();
        return body;
    }
}
